#lang racket/base

(require (only-in racket/list flatten))
(require (only-in "../aux.rkt"
           chunk stride list-ref/default make-stats stats->string))
(require (only-in "../CLOSURE/closure.rkt"
           closure register-static))
(require (prefix-in G. "../graph.rkt"))
(require (prefix-in TG. "../taskgraph.rkt"))
(require (prefix-in S. "../schedule.rkt"))
(require (only-in "../worker.rkt"
           parse-worker-args worker-args->list run-worker))
(require (only-in "../master.rkt"
           taskgraph run-schedule
           parse-master-args master-args-actual master-args->list
           start-workers workers stop-workers
           get-group-timings tm-eval tm-overhead fuse-timings/costs))
(require (only-in "../skel.rkt"
           wrap cluster par-map par-map/clustered
           par-reduce par-par-reduce/clustered par-par-map-reduce/clustered))

(define mod-id #"sumeuler")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Sum of Euler's totient function

;; Own implementation of gcd (Pycket's builtin is too slow, and might not
;; trace correctly)
(define (gcd a b)
  (if (= b 0)
      a
      (gcd b (modulo a b))))

(define (totient n)
  (for/sum ([k (in-range n)]) (if (= (gcd n k) 1) 1 0)))

;; Sum of `totient` over a sequence integers `ns`.
(define (sum-totient ns)
  (apply + (for/list ([n ns]) (totient n))))


;; Det init trace cache by calling sum-totient on fixed random list of ints
(define random-data
  '(987 633 583 261 348 128 860 333 953 482 37 329 270 601 767 977 351
    545 296 672 924 475 274 844 61 327 516 753 50 154 311 652 499 665
    40 932 22 935 828 191 69 220 196 657 167 375 783 362 662 9))
(unless (eq? (sum-totient random-data) 14574)
  (displayln "PANIC!"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; parallel version (no clustering, no fusion)

(define (par-sum-totient-1 ns)
  (define as (wrap ns))
  (define bs (par-map (closure totient/stat) as))
  (par-reduce (closure +/stat) bs))

(define totient/stat
  (register-static totient mod-id))

(define +/stat
  (register-static + mod-id))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clustered (chunking) parallel version (only map clustered, no fusion)

(define (par-sum-totient-6 ns chunksize)
  (define as (cluster (lambda (lst) (chunk lst chunksize)) (wrap ns)))
  (define bs (par-map/clustered (closure totient/stat) as))
  (par-reduce (closure +-flatten/stat) bs))

(define (+-flatten . xs)
  (apply + (flatten xs)))
;(define +-flatten
;  (compose (lambda (xs) (apply + xs)) flatten list))

(define +-flatten/stat
  (register-static +-flatten mod-id))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clustered (chunking) parallel version (no fusion)

(define (par-sum-totient-2 ns chunksize)
  (define as (cluster (lambda (lst) (chunk lst chunksize)) (wrap ns)))
  (define bs (par-map/clustered (closure totient/stat) as))
  (par-par-reduce/clustered (closure +/stat) bs))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clustered (striding) parallel version (no fusion)

(define (par-sum-totient-3 ns stridewidth)
  (define as (cluster (lambda (lst) (stride lst stridewidth)) (wrap ns)))
  (define bs (par-map/clustered (closure totient/stat) as))
  (par-par-reduce/clustered (closure +/stat) bs))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clustered (chunking) and fused parallel version

(define (par-sum-totient-4 ns chunksize)
  (define as (cluster (lambda (lst) (chunk lst chunksize)) (wrap ns)))
  (par-par-map-reduce/clustered (closure totient/stat) (closure +/stat) as))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clustered (striding) and fused parallel version

(define (par-sum-totient-5 ns stridewidth)
  (define as (cluster (lambda (lst) (stride lst stridewidth)) (wrap ns)))
  (par-par-map-reduce/clustered (closure totient/stat) (closure +/stat) as))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; run function

;; Runs `make-taskgraph` function applies to `args` in parallel;
;; returns the result and a cloned schedule (for post-mortem cost analysis)
(define (run make-taskgraph . args)
  (define result-future (apply make-taskgraph args))
  (define schedule (S.make-simple-schedule taskgraph))
  (define cloned-schedule (G.clone-graph schedule))
  (run-schedule schedule)
  (define res (TG.get-future taskgraph result-future))
  (values res cloned-schedule))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; "main" script
;;
;; Arguments (for master, after //)
;; 1) mode      [(or/c "seq" "par" "chunk" "stride" "fchunk" "fstride") = "seq"]
;; 2) lower     [number? = 6001]
;; 3) len       [number? = 4000]
;; 4) grainsize [number? = 250]
;; 5) dumpcost  [(or/c "no" "yes" "graph") = "no"]

(define all-args (vector->list (current-command-line-arguments)))
(let ([wargs (parse-worker-args all-args)])
  (if wargs
    (begin
      ;(printf "worker ~a\n" (worker-args->list wargs))
      (run-worker wargs))
    (let ([margs (parse-master-args all-args)])
      (when margs
        (define args      (master-args-actual margs))
        (define mode      (list-ref/default "seq" args 0))
        (define lower     (string->number (list-ref/default "6001" args 1)))
        (define len       (string->number (list-ref/default "4000" args 2)))
        (define grainsize (string->number (list-ref/default  "250" args 3)))
        (define dumpcost  (list-ref/default "no" args 4))
        (define ns        (for/list ([n (in-range lower (+ lower len))]) n))
        (printf "master ~a\n" (master-args->list margs))
        (start-workers margs)
        (define-values (result cloned-schedule)
          (cond
            [(equal? mode "seq")
               (printf "{seq} (sum-totient [~a..~a])\n"
                 lower (sub1 (+ lower len)))
               (time (values (sum-totient ns) (G.new-graph)))]
            [(equal? mode "par")
               (printf "{par,workers=~a} (sum-totient [~a..~a])\n"
                 (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-1 ns))]
            [(equal? mode "chunk")
               (printf "{chunk,size=~a,workers=~a} (sum-totient [~a..~a])\n"
                 grainsize (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-2 ns grainsize))]
            [(equal? mode "mchunk")
               (printf "{mchunk,size=~a,workers=~a} (sum-totient [~a..~a])\n"
                 grainsize (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-6 ns grainsize))]
            [(equal? mode "stride")
               (printf "{stride,width=~a,workers=~a} (sum-totient [~a..~a])\n"
                 grainsize (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-3 ns grainsize))]
            [(equal? mode "fchunk")
               (printf "{fchunk,size=~a,workers=~a} (sum-totient [~a..~a])\n"
                 grainsize (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-4 ns grainsize))]
            [(equal? mode "fstride")
               (printf "{fstride,width=~a,workers=~a} (sum-totient [~a..~a])\n"
                 grainsize (workers) lower (sub1 (+ lower len)))
               (time (run par-sum-totient-5 ns grainsize))]
            [else (values #f (G.new-graph))]))
        ;; print task group timing stats
        (define csched (S.compress-schedule cloned-schedule))
        (define tms (reverse (get-group-timings)))
        (printf "runtime  ~a\n"
          (stats->string (make-stats (map (compose tm-eval cdr) tms))))
        (printf "overhead ~a\n"
          (stats->string (make-stats (map (compose tm-overhead cdr) tms))))
        ;; print task cost stats
        (for ([grp (in-list (TG.get-task-costs taskgraph))])
          (printf "taskcost ~a ~a\n"
            (stats->string (make-stats (cdr grp))) (car grp)))
        ;; dump timings fused with costs (by task group, in order of completion)
        (when (or (equal? dumpcost "yes") (equal? dumpcost "graph"))
          (define tmcs (fuse-timings/costs csched tms taskgraph))
          (printf "GROUP key t-eval t-overhead task-costs\n")
          (for ([k-tm-cost (in-list tmcs)])
            (define k          (car k-tm-cost))
            (define t-eval     (tm-eval (cadr k-tm-cost)))
            (define t-overhead (tm-overhead (cadr k-tm-cost)))
            (define costs      (cddr k-tm-cost))
            (printf "group ~a ~a ~a ~a\n" k t-eval t-overhead costs)))
        (when (equal? dumpcost "graph")
          (TG.print-graph taskgraph)
          (printf "(define compressed-schedule '~a)\n" csched)
          (printf "(define serialized-size '~a)\n"
            (S.size-profile-schedule cloned-schedule taskgraph)))
        (printf "~a\n" result)
        (stop-workers)))))
